/* irc.c - IRC interface */

#define PREFIX_COMMAND_CHAR '!'

#define IRCMSG(msg) irc_cmd_msg(session, creds.channel, msg)

typedef struct
{
  char *  username;
  char *  password;
  char *  channel;
  char *  server;
  int     port;
} creds_t;

typedef struct
{
  const char * key;
  int val;
  UT_hash_handle hh;
} map_t;

VARDECL creds_t creds =
{
  .username = NULL,
  .password = NULL,
  .channel  = NULL,
  .server   = NULL, /* localhost? */
#ifndef IRC_SSL_SUPPORT
  .port     = 6667
#else
  .port     = 6669
#endif /* !IRC_SSL_SUPPORT */
};

VARDECL map_t * map = NULL;

VARDECL char * ident_password = NULL;

VARDECL irc_session_t * session;
VARDECL irc_callbacks_t callbacks;

VARDECL char * current_username = NULL;

VARDECL char const * help_msg =
/* IRC_GREEN "!help           " IRC_STOP "   : This message\n" */
IRC_GREEN "!remind         " IRC_STOP "   : Dump current assignment\n"
IRC_GREEN "!reroll         " IRC_STOP "   : Rerolls assignment\n"
/* IRC_GREEN "!repo <link>" IRC_STOP "   : Sets project repository link\n" */
/* IRC_GREEN "!raw <sql>      " IRC_STOP "   : Execute raw SQL\n" */
/* IRC_GREEN "!dump           " IRC_STOP "   : List all possible projects\n" */
/* IRC_GREEN "!request        " IRC_STOP "   : Request personal project\n" */
IRC_GREEN "!remind         " IRC_STOP "   : Prints your assignment\n"
IRC_GREEN "!poll           " IRC_STOP "   : Start the democratic process (N message ...])\n"
IRC_GREEN "!y              " IRC_STOP "   : Votes yes\n"
IRC_GREEN "!n              " IRC_STOP "   : Votes no\n"
IRC_GREEN "!stop           " IRC_STOP "   : Stop the current polling\n"
IRC_BLUE  "!magic          " IRC_STOP "   : (N) Who knows.\n"
IRC_GREEN "!dice           " IRC_STOP "   : (N or AdX format) Rolls dice.";

DECL void parse_command(char * cmd);
#ifdef SECONDARY_NAMES_BOT
DECL size_t parse_secondary_username(char const * msg);
#endif /* SECONDARY_NAMES_BOT */

DECL char *
get_username(const char * origin)
{
  const char USERNAME_TERMINATOR = '!';
  int i = 0;
  char * r;
  while (origin[i] != USERNAME_TERMINATOR)
  { i++; }
  r = (char *) malloc(i + 1);
  if (r)
  {
    strncpy(r, origin, i);
    r[i] = '\00';
  }
  return r;
}

DECL void
ircmsg(char const * reciever, char const * fmt,
       ...)
{
  va_list args;
  char * fmtdmsg;
  char * swp;
  const char * delim = "\n";
  char * data;
  if(!strcmp(fmt, "") || fmt == NULL)
  { return; }

  va_start(args, fmt);
  if(vasprintf(&fmtdmsg, fmt, args) == -1)
  { exit(1); }

  puts(fmtdmsg);
  data = strtok(fmtdmsg, delim);
  do
  {
    swp = irc_color_convert_to_mirc(data);
    irc_cmd_msg(session, reciever, swp);
    free(swp);
  } while((data = strtok(NULL, delim), data));

  free(fmtdmsg);
  va_end(args);
}



DECL void
event_connect(irc_session_t * lsession,
              char const  * event,
              char const  * origin,
              char const ** params,
              unsigned int count)
{
  (void) event;
  (void) origin;
  (void) params;
  (void) count;
  /* msg ChanServ IDENTIFY? */
  irc_cmd_join(lsession, creds.channel, 0);
  if (ident_password)
  {
    ircmsg("NickServ", "IDENTIFY %s", ident_password);
    memset(ident_password, '\0', strlen(ident_password));
  }

#ifdef INITIAL_ASSIGNMENT_MESSAGE
  if(is_no_assignment(creds.channel))
  {
    ircmsg(creds.channel,
					 IRC_RED "No assignment for this channel. Finding a new..." IRC_STOP);
    random_assign(creds.channel);
  }
  ircmsg(creds.channel, remind(creds.channel));
#endif /* INITIAL_ASSIGNMENT_MESSAGE */
}

DECL void
event_channel(irc_session_t * lsession,
              char  * event,
              char  * origin,
              char ** params,
              unsigned int count)
{
  /* char const * channel = params[0]; */
  char * message = params[1];
  (void) lsession;
  (void) event;
  (void) origin;
  (void) message;
  (void) count;
  /* fetches username and logs message */
	if (!(current_username = get_username(origin)))
  { return; }
  printf("<%s> %s\n", current_username, message);
  /* Detects any respecified names */
#ifdef SECONDARY_NAMES_BOT
  message += parse_secondary_username(message);
#endif /* SECONDARY_NAMES_BOT */
  /* Parse commands */
  if (*message == PREFIX_COMMAND_CHAR &&
      message[1] != '\0')
  { parse_command(message+1); }
  free(current_username);
  current_username = NULL;
}

int name_sort(const map_t * a,
              const map_t * b)
{
    return strcmp(a->key, b->key);
}

DECL void
parse_command_init(void)
{
  size_t i;
  static const char * commands[] =
  {
    "n",      /* 0  */
    "remind", /* 1  */
    "y",      /* 2  */
    "help",   /* 3  */
    "magic",  /* 4  */
    "dice",   /* 5  */
    "reroll", /* 6  */
    "poll",   /* 7  */
    "stop",   /* 8  */
    NULL
  };
  map_t * s;
  for (i = 0; commands[i]; ++i)
  {
    s = (map_t *) malloc(sizeof(map_t));
    s->key = strdup(commands[i]);
    s->val = i;
    HASH_ADD_STR(map, key, s);
  }
  HASH_SORT(map, name_sort);
}

DECL void
parse_command_free(void)
{
  map_t * s, * tmp;
  HASH_ITER(hh, map, s, tmp) {
      HASH_DEL(map, s);
      free((void*)s->key);
      free(s);
  }
}

#define DICE_DEFAULT 6
#define ARGMASK 16

DECL int
has_arg(char const * cmd)
{
  char const * start = cmd;
  while (isalnum(*cmd))
  {
    if (*cmd == '\0')
    { return 0; }
    ++cmd;
  }
  while (*cmd != '\0')
  {
    if (isspace(*cmd))
    { return cmd - start; }
    ++cmd;
  }
  return 0;
}

DECL void
parse_command(char * cmd)
{
  static int vote_count, vote;
  int dice = DICE_DEFAULT;
  size_t len = has_arg(cmd);
  size_t i = 0;
  char x;
  char * arg = NULL;
  char * msgswp;
  map_t * s;

  /* x = cmd[len]; */
  /* fprintf(stderr, "%ld\n", len); */
  /* cmd[len] = cmd[len] != '\0' ? '\0' : cmd[len]; */
  /* HASH_FIND_STR(map, cmd, s); */
  /* fprintf(stderr, "cmd: "); */
  /* while (cmd[i]) */
  /* { */
  /*   x = cmd[len]; */
  /*   cmd[len] = cmd[len] ? cmd[len] : '\0'; */
  /*   fprintf(stderr, "%c:%d ", cmd[i], cmd[i]); */
  /*   ++i; */
  /* } */
  /* cmd[len] = x; */
  /* putchar('\n'); */

  x = cmd[len];
  if (len &&
      cmd[len] != 0)
  { cmd[len] = '\0'; }
  HASH_FIND_STR(map, cmd, s);
  cmd[len] = x;
  if (cmd[len])
  {
    arg = cmd + len;
  }
  if (arg) { fprintf(stderr, "arg: '%s'\n", arg); }

  if (s)
  {
    if (arg) {s->val |= ARGMASK; }
    switch (s->val)
    {
    case 0 : /* n      */
    case 2 : /* y      */
      vote = s->val - 1; /* I had to, man */
      if (--vote_count)
      { ircmsg(creds.channel, "Votes remaining: %d", vote_count); }
      else
      { ircmsg(creds.channel,
               "poll: %s",
               vote == 0 ? "UNDECIDED" :
               vote  > 0 ?    "PASSED" :
               "REJECTED"); }
      break;
    case 1 :  /* remind */
      msgswp = remind(current_username);
      ircmsg(creds.channel, "%s: %s", current_username, msgswp);
      break;
    case 3 :  /* help   */
      ircmsg(creds.channel, help_msg);
      break;
    case 4 :  /* magic  */
      break;
    case 5 | 16 :
      dice = atoi(arg);
      /* fallthrough */
    case 5 :  /* dice   */
      ircmsg(creds.channel, "%s: " IRC_YELLOW "%d" IRC_STOP, current_username, (rand() % dice) + 1);
      break;
    case 6 :  /* reroll */
      purge_assignments(current_username);
      random_assign(current_username);
      ircmsg(creds.channel, "%s: %s", current_username, remind(current_username));
      break;
    case 7 | ARGMASK :  /* poll   */
      vote = 0;
      vote_count = atoi(arg);
      if (!vote_count)
      case 7 : { ircmsg(creds.channel, "!poll NUMBER_OF_VOTES, What you're voting on"); }
      else
      { ircmsg(creds.channel, "poll: start, vote with !y or !n"); }
      break;
    case 8 :  /* stop   */
      if (vote_count)
      {
        ircmsg(creds.channel, "poll: stop");
        vote_count = 0;
      }
      break;
    default:
      fprintf(stderr, "Unhandled match.\n");
    }
  }
}
